package com.tzld.piaoquan.recommend.server.service.score.model;


import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;



public class ModelManager {
    private static final int SCHEDULE_PERIOD = 10;
    private static final Logger LOGGER = LoggerFactory.getLogger(ModelManager.class);
    private static final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);
    private static final String SUCCESS = "_SUCCESS";
    private static ModelManager instance;
    Map<String, ModelLoadTask> loadTasks = new HashMap<String, ModelLoadTask>();
    Map<String, String> modelPathMap = new HashMap<String, String>();
    private Config config = ConfigFactory.load("hdfs_conf.properties");
    private Configuration hdfsConf = new Configuration();

    private ModelManager() {
        //配置不同环境的hdfs conf
        String coreSiteFile = config.hasPath("hdfs.coreSiteFile") ? StringUtils.trim(config.getString("hdfs.coreSiteFile")) : "core-site.xml";
        String hdfsSiteFile = config.hasPath("hdfs.hdfsSiteFile") ? StringUtils.trim(config.getString("hdfs.hdfsSiteFile")) : "hdfs-site.xml";
        hdfsConf.addResource(coreSiteFile);
        hdfsConf.addResource(hdfsSiteFile);

        start(SCHEDULE_PERIOD);
    }

    public static ModelManager getInstance() {
        if (instance == null) {
            synchronized (ModelManager.class) {
                if (instance == null) {
                    instance = new ModelManager();
                }
            }
        }
        return instance;
    }

    /**
     * 添加一个加载任务到管理器
     *
     * @param modelName  Model的名字, 注册到ModelManager的不同model需要不同的名字
     * @param path       Model在HDFS上的路径
     * @param modelClass Model的子类型
     */
    public void registerModel(String modelName, String path, Class<? extends Model> modelClass) throws ModelRegisterException, IOException {
        if (modelPathMap.containsKey(modelName)) {
            String oldPath = modelPathMap.get(modelName);
            if (path.equals(oldPath)) {
                //如果模型的path没有发生改变, 不做任何操作
                LOGGER.info("Model [{}] and Path [{}] has exist", modelName, path);
                return;
            } else {
                //如果模型的path发生改变, 需要注销掉原有的任务
                unRegisterModel(modelName);
            }
        }

        modelPathMap.put(modelName, path);
        if (loadTasks.containsKey(path)) {
            ModelLoadTask loadTask = loadTasks.get(path);
            loadTask.refCount++;
        } else {
            ModelLoadTask task = new ModelLoadTask(path, modelClass);
            task.refCount++;
            loadTasks.put(path, task);
            loadModel(task, false, true);
        }
    }

    /**
     * 删除一个加载任务
     *
     * @param modelName Model的名字, 需要和registerModel的名字一致
     */
    private void unRegisterModel(String modelName) {
        if (modelPathMap.containsKey(modelName)) {
            String path = modelPathMap.get(modelName);
            if (loadTasks.containsKey(path)) {
                ModelLoadTask task = loadTasks.get(path);
                task.refCount--;
                if (task.refCount == 0) {
                    loadTasks.remove(path);
                }
            }
            modelPathMap.remove(modelName);
        }
    }

    /**
     * @param modelName
     * @return
     */
    public Model getModel(String modelName) {
        if (modelPathMap.containsKey(modelName) && loadTasks.containsKey(modelPathMap.get(modelName))) {
            return loadTasks.get(modelPathMap.get(modelName)).model;
        } else {
            return null;
        }
    }

    /**
     * 开始调度
     *
     * @param period
     */
    protected void start(long period) {
        final Runnable task = new Runnable() {
            public void run() {
                // 模型更新开关
                // boolean modelUpdateSwitch = Configuration.getBoolean("recommend-service-framework.model-update-switch", true);
                boolean modelUpdateSwitch =  true;
                LOGGER.info("model update switch [{}]", modelUpdateSwitch);
                if (modelUpdateSwitch) {
                    updateModels(false);
                }
            }
        };
        scheduler.scheduleAtFixedRate(task, 10, period, TimeUnit.MINUTES); // 10分钟
    }

    /**
     * 更新模型
     */
    public void updateModels(final boolean isForceLoads) {
        LOGGER.info("begin to update: [{}]", loadTasks.keySet().size());
        for (String modelPath : loadTasks.keySet()) {
            LOGGER.debug("loadtask model path [{}]", modelPath);
            ModelLoadTask task = loadTasks.get(modelPath);
            loadModel(task, isForceLoads, false);
        }
    }

    /**
     * 检查并加载模型
     *
     * @param loadTask
     */
    private void loadModel(final ModelLoadTask loadTask, final boolean isForceLoads, final boolean isRegister) {
        if (loadTask.isLoading) {
            return;
        }
        FSDataInputStream fin = null;
        try {
            loadTask.isLoading = true;
            Path path = new Path(loadTask.path);
            hdfsConf.setBoolean("fs.hdfs.impl.disable.cache", true);

            //根据path选择相应的filesystem
            FileSystem fs = path.getFileSystem(hdfsConf);
            FileStatus fileStatus = fs.getFileStatus(path);

            // model is file, modify sign is file ModificationTime
            if (fileStatus.isFile()) {

                long timeStamp = fileStatus.getModificationTime();
                LOGGER.debug("model [{}] last update time: [{}]", loadTask.path, loadTask.lastModifyTime + ":" + timeStamp);
                if (loadTask.lastModifyTime != timeStamp || isForceLoads) {
                    LOGGER.info("find model [{}] file changed, ready to update, last modify: [{}]", loadTask.path, loadTask.lastModifyTime);

                    Path modelFile = new Path(loadTask.path);
                    fin = fs.open(modelFile);
                    Model model = loadTask.modelClass.newInstance();
                    if (model.loadFromStream(new InputStreamReader(fin))) {
                        loadTask.model = model;
                        loadTask.lastModifyTime = timeStamp;
                        //LOGGER.info("finish load model: [{}], [{}]", loadTask.path);
                        LOGGER.info("update model: [{}], [{}]", loadTask.path, loadTask.lastModifyTime);
                    } else {
                        LOGGER.error("load model fail: [{}]", loadTask.path);
                    }
                }
                return;
            }

            // model is dir, modify sign is file _SUCCESS file ModificationTime
            if (fileStatus.isDirectory()) {

                String successFilePath = String.format("%s/%s", loadTask.path, SUCCESS);
                Path successPath = new Path(successFilePath);
                long currSuccFileModifyTime = 0;
                if (!isForceLoads) {
                    try {
                        FileSystem fsSucc = successPath.getFileSystem(hdfsConf);
                        FileStatus fileStatusSucc = fsSucc.getFileStatus(successPath);
                        currSuccFileModifyTime = fileStatusSucc.getModificationTime();

                        //check model whether update
                        if (loadTask.lastModifyTime == currSuccFileModifyTime) {
                            LOGGER.info("[INFO MESSAGE] _SUCCESS modifytime is not changed, last update time [{}]," +
                                    "current modify time [{}], model path: [{}]", new Object[]{loadTask.lastModifyTime,
                                    currSuccFileModifyTime, loadTask.path});
                            return;
                        }
                    } catch (Exception e) {
                        LOGGER.error("check _SUCCESS update time occur error, _SUCCESS path [{}], Exception: [{}]",
                                successPath, ExceptionUtils.getFullStackTrace(e));
                        return;
                    }
                }

                // success has modify, multi thread load model
                Model model = loadTask.modelClass.newInstance();
                if (model.loadPartitions(loadTask.path, isRegister)) {
                    loadTask.model = model;
                    loadTask.lastModifyTime = currSuccFileModifyTime;

                    LOGGER.info("[SUCCESS] model load, update time [{}], model size [{}], model path [{}]",
                            new Object[]{loadTask.lastModifyTime, model.getModelSize(), loadTask.path});
                } else {
                    LOGGER.error("load model [failed], model path [{}]", loadTask.path);
                }
            }

        } catch (Exception e) {
            LOGGER.error("update model fail:[{}], [{}]", loadTask.path, ExceptionUtils.getFullStackTrace(e));
        } finally {
            loadTask.isLoading = false;
            if (fin != null) {
                try {
                    fin.close();
                } catch (IOException e) {
                    LOGGER.error("close [{}] fail: [{}]", loadTask.path, ExceptionUtils.getFullStackTrace(e));
                }
            }
        }
    }

    public class ModelRegisterException extends Exception {

        public ModelRegisterException(String s) {
            super(s);
        }
    }

    /**
     * 调度的任务单元
     */
    private class ModelLoadTask {

        private int refCount;
        private String path;
        private long lastModifyTime;
        private boolean isLoading;
        private Class<? extends Model> modelClass;
        private Model model;

        ModelLoadTask(String path, Class<? extends Model> modelClass) {
            this.refCount = 0;
            this.path = path;
            this.lastModifyTime = 0;
            this.modelClass = modelClass;
        }
    }
}
